#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Plot amplitude modulated tone.

"""

from __future__ import division, absolute_import, print_function

import os
import numpy as np
import matplotlib.pyplot as plt

import thorns.waves as wv

def main():

    fs = 48e3

    sound = wv.amplitude_modulated_tone(
        fs=fs,
        fm=100,
        fc=1e3,
        m=0.7,
        duration=0.1,
    )

    wv.plot_signal(sound, fs=fs)

    wv.show()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module implements permanent store for data.

"""
from __future__ import division, print_function, absolute_import
from __future__ import unicode_literals

import os
import datetime
import logging

from sys import version
import shelve
import collections
import cPickle as pickle

import tables

import numpy as np
import pandas as pd

from math import isnan

logger = logging.getLogger('thorns')


class NDArrayHandler(object):
    @staticmethod
    def tag(a):
        return 'array'

    @staticmethod
    def rep(a):
        return a.tolist()

    @staticmethod
    def string_rep(a):
        return None


class NumpyIntHandler(object):
    @staticmethod
    def tag(i):
        return 'i'

    @staticmethod
    def rep(i):
        return int(i)

    @staticmethod
    def string_rep(i):
        return str(i)


class NumpyFloatHandler(object):
    @staticmethod
    def tag(f):
        return "z" if isnan(f) or f in (float('Inf'), float('-Inf')) else "d"

    @staticmethod
    def rep(f):
        if isnan(f):
            return "NaN"
        if f == float('Inf'):
            return "INF"
        if f == float("-Inf"):
            return "-INF"
        return float(f)

    @staticmethod
    def string_rep(f):
        return str(f)


def get_store(name='store', workdir='work'):
    """Return a quick and dirty shelve based persisten dict-like store.

    """
    fname = os.path.join(workdir, name + '.db')

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    store = shelve.open(fname, protocol=-1)

    return store


def dump(data, name, workdir='work', backend='pickle'):
    """Dump a data object."""
    if not os.path.exists(workdir):
        os.makedirs(workdir)

    if backend == 'pickle':
        fname = os.path.join(workdir, name+'.pkl')

        with open(fname, 'w') as f:
            pickle.dump(data, f, -1)

    elif backend == 'transit':
        fname = os.path.join(workdir, name+'.json')

        from transit.writer import Writer

        with open(fname, 'w') as f:
            writer = Writer(f, 'json')

            writer.register(np.ndarray, NDArrayHandler)
            writer.register(np.int64, NumpyIntHandler)
            writer.register(np.float64, NumpyFloatHandler)

            writer.write(data)



def dumpdb(data, name='dump', workdir='work', backend='hdf', kwargs=None):
    """Dump data in order to recall the most up-to-date records later.

    Parameters
    ----------
    data : pd.DataFrame
        Data that will be appended to the database.
    name : str, optional
        Base name of the pickle file.
    workdir : str, optional
        Directory for the data.
    backend : str, optional
        Data storage format.
    kwargs : dict, optional
        Additional parameters common for all data (MultiIndex will be
        extended).

    """
    if not os.path.exists(workdir):
        os.makedirs(workdir)

    if kwargs is not None:
        for k,v in kwargs.items():
            data[k] = v
        data = data.set_index(kwargs.keys(), append=True)

    if backend == 'hdf':
        fname = os.path.join(workdir, name+'.h5')

        logger.info("Dumping data into {}.".format(fname))

        now = datetime.datetime.now()
        key = now.strftime("T%Y%m%d_%H%M%S_%f")

        store = pd.io.pytables.HDFStore(fname, 'a')
        store[key] = data

        store.close()

    elif backend == 'shelve':
        fname = os.path.join(workdir, name+'.db')

        logger.info("Dumping data into {}.".format(fname))

        now = datetime.datetime.now()
        key = now.strftime("T%Y%m%d_%H%M%S_%f")

        store = shelve.open(fname, protocol=-1)

        store[key] = data

        store.close()

    elif backend == 'transit':
        from transit.writer import Writer

        logger.info("Dumping data into {}.".format(fname))

        d = data.reset_index().to_dict('records')

        dump(d, name=name, workdir=workdir, backend='transit')

    else:
        raise NotImplementedError()



def loaddb(name='dump', workdir='work', backend='hdf', timestamp=False, load_all=False):
    """Recall dumped data discarding duplicated records.

    Parameters
    ----------
    name : str, optional
        Base of the data filename.
    workdir : str, optional
        Directory where the data is stored.
    backend : str, optional
        Backend used for data storage.
    timestamp : bool, optional
        Add an extra column with timestamps to the index.
    load_all : bool, optional
        If True, data from all experiments will be loaded from the
        dumpdb file.  The default is to load only the most recent
        data.

    Returns
    -------
    pd.DataFrame
        Data without duplicates.

    """
    if timestamp:
        raise NotImplementedError("Should add extra columnt with timestamps to the index of the output.")


    if backend == 'hdf':
        db = _loaddb_hdf(name, workdir, load_all)
    elif backend == 'shelve':
        db = _loaddb_shelve(name, workdir, load_all)
    else:
        raise NotImplementedError()

    return db



def _loaddb_hdf(name, workdir, load_all):

    fname = os.path.join(workdir, name+'.h5')
    store = pd.io.pytables.HDFStore(fname, 'r')

    logger.info("Loading data from {}".format(fname))

    if load_all:
        xkeys = collections.OrderedDict() # poor-man's ordered set
        dbs = []

        ### Get all tables from the store
        for t in sorted(store.keys()):
            df = store[t]

            # Just want ordered unique values in xkeys (ordered set would
            # simplify it: orderedset.update(df.index.names))
            for name in df.index.names:
                xkeys.setdefault(name)

            df = df.reset_index()
            dbs.append(df)

        db = pd.concat(dbs)

    else:
        last_key = sorted(store.keys())[-1]
        df = store[last_key]

        xkeys = df.index.names

        db = df.reset_index()

    store.close()

    ### Drop duplicates and set index
    db = db.drop_duplicates(
        subset=list(xkeys),
        keep='last',
    )

    db = db.set_index(list(xkeys))

    return db




def _loaddb_shelve(name, workdir, load_all):

    if load_all:
        raise NotImplementedError()

    fname = os.path.join(workdir, name+'.db')
    store = shelve.open(fname, protocol=-1)

    logger.info("Loading data from {}".format(fname))

    last_key = sorted(store.keys())[-1]

    df = store[last_key]

    store.close()

    # Drop records with duplicated keys
    xkeys = df.index.names

    db = df.reset_index()

    db = db.drop_duplicates(
        subset=list(xkeys),
        keep='last',
    )

    db = db.set_index(list(xkeys))

    return db



# -*- coding: utf-8 -*-

"""Sample spike trasins.

"""

from __future__ import division, absolute_import, print_function

from os.path import join, dirname

import numpy as np
import pandas as pd


def load_anf_zilany2014():
    """Load and return sample auditory nerve spike trains generated with
    Zilany et al. (2014) inner ear model.

    The simulus was a 50 ms ramped pure tone at 50 dB_SPL (without
    padding).


    Returns
    -------
    spike_trains
        Responses of high-spontaneous rate auditory nerve fibers with
        the characteristic ferquency (CF) of 1 kHz.

    """

    module_path = dirname(__file__)

    spike_trains = pd.read_pickle(
        join(module_path, "anf_zilany2014.pkl")
    )

    return spike_trains



def main():
    pass

if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import division, print_function, absolute_import

from numpy.testing import assert_equal
from pandas.util.testing import assert_frame_equal

import numpy as np
import pandas as pd

import thorns as th



def test_from_array():

    fs = 1e3
    a = np.array(
        [[0,0,0,1,0,0],
         [0,2,1,0,0,0]]
    ).T

    result = th.make_trains(a, fs=fs)


    expected = th.make_trains(
        [np.array([3])/fs,
         np.array([1,1,2])/fs],
        duration = 6/fs
    )


    assert_frame_equal(
        expected,
        result
    )




def test_from_arrays():

    arrays = [
        [1,2,3],
        [1,3,5,7],
        [0,8]
    ]


    expected = {
        'spikes': arrays,
        'duration': np.repeat(10., len(arrays))
    }
    expected = pd.DataFrame(expected)





    result = th.spikes._arrays_to_trains(
        arrays,
        duration=10.0
    )

    assert_frame_equal(
        result,
        expected
    )




    result = th.make_trains(
        arrays,
        duration=10.0
    )
    assert_frame_equal(
        expected,
        result
    )




def test_make_empty_trains():

    spikes = [[], []]

    trains = th.make_trains(
        spikes
    )

    expected = pd.DataFrame({
        'spikes': spikes,
        'duration': np.repeat(0., len(spikes))
    })

    assert_frame_equal(
        trains,
        expected
    )






    trains = th.make_trains(
        spikes,
        duration=10
    )

    expected = pd.DataFrame({
        'spikes': spikes,
        'duration': np.repeat(10., len(spikes))
    })

    assert_frame_equal(
        trains,
        expected
    )





def test_trains_to_array():


    trains = th.make_trains(
        [[1], [2,3]],
        duration=5
    )

    result = th.spikes.trains_to_array(
        trains,
        fs=1
    )


    assert_equal(
        result,
        [[0,0],
         [1,0],
         [0,1],
         [0,1],
         [0,0]]
    )






def test_accumulate():

    trains = th.make_trains(
        [[1], [2], [3], []],
        cfs=[2,1,2,3]
    )

    accumulated = th.accumulate(trains)

    expected = th.make_trains(
        [[2], [1,3], []],
        cfs=[1,2,3]
    )

    assert_frame_equal(accumulated, expected)




def test_select_trains():

    trains = th.make_trains(
        [[1], [2], [3], [4]],
        duration=4,
        cfs=[0,0,1,1],
        idx=[0,1,0,1]
    )



    selected = th.select_trains(
        trains,
        cfs=1,
        idx=1
    )

    expected = th.make_trains(
        [[4]],
        duration=4,
        cfs=[1],
        idx=[1]
    )
    expected.index = [3]

    assert_frame_equal(
        selected,
        expected,
    )




def test_trim():

    trains = th.make_trains(
        [[1,2,3,4],
         [3,4,5,6]],
        duration=8,
        type='hsr'
    )


    trimmed = th.trim(trains, 2, 4)


    expected = th.make_trains(
        [[0,1,2], [1,2]],
        duration=2,
        type='hsr'
    )


    assert_frame_equal(
        trimmed,
        expected
    )



def test_fold():
    trains = th.make_trains(
        [[1.1, 2.1, 3.1], [2.1, 3.1, 5.1]],
        duration=6.5,
        cf=[1,2]
    )

    folded = th.fold(trains, 1)


    expected = th.make_trains(
        [[  ], [.1], [.1], [.1], [  ], [  ], [  ],
         [  ], [  ], [.1], [.1], [  ], [.1], [  ]],
        duration=[1, 1, 1, 1, 1, 1, 0.5,
                  1, 1, 1, 1, 1, 1, 0.5],
        cf=[1, 1, 1, 1, 1, 1, 1,
            2, 2, 2, 2, 2, 2, 2]
    )

    assert_frame_equal(
        folded,
        expected,
    )





def test_trim():

    trains = th.make_trains(
        [[1, 2], [2, 3]],
        duration=10,
        cf=1e3
    )

    trimmed = th.trim(trains, 1.5, 2.5)

    expected = th.make_trains(
        [[0.5], [0.5]],
        duration=1,
        cf=1e3
    )


    assert_frame_equal(trimmed, expected)




def test_trim_without_stop():

    trains = th.make_trains(
        [[1, 2], [2, 3]],
        duration=10,
        cf=1e3
    )

    trimmed = th.trim(trains, 2)

    expected = th.make_trains(
        [[0], [0, 1]],
        duration=8,
        cf=1e3
    )

    assert_frame_equal(trimmed, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This demo generates sample periodic spike trains and plots a
period histogram.

"""

from __future__ import division, absolute_import, print_function

import numpy as np

import thorns as th



def main():

    fs = 10e3                   # Hz
    freq = 10                   # Hz


    ### Calculate spike probability over time (we want to have a
    ### periodic function for a pretty plot)
    t = np.arange(0, 1, 1/fs)
    spike_probability = 0.2 * np.abs(np.sin(2 * np.pi * freq * t))
    rand = np.random.rand(len(spike_probability))




    ### Generate spike trains
    spike_array = np.zeros_like(spike_probability)
    spike_array[rand < spike_probability] = 1
    spike_array = np.expand_dims(spike_array, axis=1)

    spike_trains = th.make_trains(
        spike_array,
        fs=fs,
    )




    ### Plot the histogram
    th.plot_period_histogram(
        spike_trains,
        freq=freq,
        nbins=128,
        density=True
    )

    th.show()






if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Map demo.

"""

from __future__ import division, print_function, absolute_import
from __future__ import unicode_literals

import numpy as np
import thorns as th


def multiply(a, b):
    """Multiply two numbers."""

    # The output should be wrapped in a dict
    y = {'y': a * b}

    return y


def main():

    # Keys in the input dict MUST correspond to the key word
    # arguments of the function (e.g. multiply)
    xs = {
        'a': np.arange(3),
        'b': np.arange(3)
    }

    # Uncomment `backend` for parallel execution
    ys = th.util.map(
        multiply,
        xs,
        # backend='multiprocessing',
    )

    print(ys)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Test nesting of thorns.unil.map

"""

from __future__ import division, print_function, absolute_import
from __future__ import unicode_literals

import tempfile
import shutil
import pytest

import numpy as np
from numpy.testing import assert_equal

import pandas as pd
from pandas.util.testing import assert_frame_equal

import thorns as th

import time


@pytest.fixture(scope="function")
def workdir(request):

    workdir = tempfile.mkdtemp()

    def fin():
        print("Removing temp dir: {}".format(workdir))

        shutil.rmtree(workdir, ignore_errors=True)

    request.addfinalizer(fin)

    return workdir





def square(x):
    return x**2



def sum_of_squares(length, workdir):

    vec = {'x': np.arange(length)}

    squares = th.util.map(
        square,
        vec,
        backend='multiprocessing', # should be inhibited to 'serial'
        cache='no',
        workdir=workdir
    )

    result = np.sum(squares.values)

    return result





def test_map_nested(workdir):

    lengths = {'length': [2, 3]}

    actual = th.util.map(
        sum_of_squares,
        lengths,
        backend='multiprocessing',
        cache='no',
        workdir=workdir,
        kwargs={'workdir': workdir},
    )

    expected = pd.DataFrame({
            'length': [2,3],
            0: [1,5],
    }).set_index('length')


    assert_frame_equal(expected, actual)
